#include "control.h"
#include "MyData.h"

/* hier stehen die verschiedenen Funktionen, mit denen der Hamster gesteuert wird */

void take(Control *c) {                                                /* Korn vom aktuellen Feld nehmen */
  HAMSTER *hms = c->myhamster->hamster;
  struct MyData *md = c->position.item->node;
  
  while ((hms_load(hms)!=HMS_MAXLOAD) && (hms_corn(hms)!=0) ) {        /* solange der Hamster nicht voll beladen ist, nimm ein Korn und verringere die Kornzahl auf */
    hms_take(hms, 1);                                                  /*dem Listenfeldum 1 */                              
    --md->corn;                                  
  }
}


void go_step(Control *c) {                                             /* einen Schritt in die aktuelle Richtung */
  HAMSTER *hms = c->myhamster->hamster;
  hms_move(hms);
}


void go_dir(Control *c, enum Direction d) {                            /* der Hamster geht einen Schritt in die angegebene Richtung */
  switch(d) {                                                          /* Auswahl der speziellen Funktionen, um in eine spezielle Richtung zu gehen, je nach dem, */
  case north:                                                          /* welche Richtung uebergeben wurde, */
    go(go_north, c);
    break;
  case south:
    go(go_south, c);
    break;
  case east:
    go(go_east, c);
    break;
  case west:
    go(go_west, c);
    break;
  default:
  }
}


void go_north(Control *c) {                                            /* ein Feld nach Norden gehen */
  turn(turn_north, c);                                                 /* Hamster nach norden drehen */
  go_step(c);                                                          /* einen Schritt machen */
}


void go_south(Control *c) {                                            /* wie go_north(Control *c), nur nach Sueden */
  turn(turn_south, c);
  go_step(c);
}


void go_west(Control *c) {                                             /* wie go_north(Control *), nur nach Westen */
  turn(turn_west, c);
  go_step(c);
}


void go_east(Control *c) {                                             /* wie go_north(Control *), nur nach Osten */
  turn(turn_east, c);
  go_step(c);
}


void go_next(Control *c) {                                             /* auf das naechste Feld in der Liste gehen (vom Startfeld weg) */
  struct MyData *md = c->position.item->node;
  go_dir(c, md->godir);                                                /* bei go_dir  wird als Parameter die in dem Listenfeld festgelegte (Hin)Richtung uebergeben */
  c->position = next(c->position);                                     /* Zeiger in der Liste nachfuehren */
}


void go_back(Control *c) {                                             /*1 feld zurueck in Richtung Startfeld */
  struct MyData *md = c->position.item->node;
  go_dir(c, md->fromdir);                                              /* bei go_dir  wird als Parameter die in dem Listenfeld festgelegte (Rueck)Richtung uebergeben */
    c->position = prev(c->position);                                   /* Zeiger in der Liste nachfuehren */
}


void go_start(Control *c) {                                            /* der Hamster laeuft (von einem beliebigen Feld, muss aber in der Liste sein) zum Startfeld */
  HAMSTER *hms = c->myhamster->hamster;
  while (c->position.item != begin(&c->myhamster->list).item) {        /* solange die aktuelle Position des Hamsters nicht das Startfeld ist, */
    take(c);                                                           /* versuche Korn zu nehmen */
    go_back(c);                                                        /* und gehe einen Schritt richtung Bau */
  }
  hms_drop(hms, HMS_MAXLOAD);                                          /* beim Bau angekommen, lege alles Korn ab */
}

 
void go_end(Control *c) {                                              /* auf das zuletzt erforschte Feld / letztes Feld in der Liste gehen */
  while (c->position.item != prev(end(&c->myhamster->list)).item) {    /* solange der Hamster nicht auf dem letzten Feld ist, */
    go_next(c);                                                        /* gehe aufs naechste Feld */
  }
}


void go(void (*f)(Control *), Control *c) {                            /* dieser Funktion werden die gehe... Funktionen und der Hamster als Parameter uebergeben */
  (*f)(c);
}


void turn_to_dir(Control *c, int e) {                                  /* allgemeines drehen des Hamsters in eine angegebene Richtung */
                                                                       /*
                                                                          e... Richtung, in die der Hamster gedreht werden soll
                                                                       */
   switch (look_dir(c)-e) {
   case 0:                                                             /* keine Drehung */
     break;
   case -3:
   case 1:                                                             /* Drehung nach rechts (z.B.: N->O) */ 
    turn(turn_right, c);
    break;
   case 3:
   case -1:                                                            /* Drehung nach links (z.B.: N->W) */
    turn(turn_left, c);
    break;
   default:                                                            /* Drehung in die entgegenges. Richtung (z.B.: N->S) */
    turn(turn_right, c);
    turn(turn_right, c);
   }
}



void turn_north(Control *c) {                                          /* Hamster nach Norden drehen */
  turn_to_dir(c,HMS_NORTH);
}


void turn_south(Control *c) {                                          /* Hamster nach Sueden drehen */
  turn_to_dir(c,HMS_SOUTH);
}


void turn_west(Control *c) {                                           /* Hamster nach Westen drehen */
  turn_to_dir(c,HMS_WEST);
}


void turn_east(Control *c) {                                           /* Hamster nach Osten drehen */
  turn_to_dir(c,HMS_EAST);
}


void turn_left(Control *c) {                                           /* Hamster nach links drehen */
  HAMSTER *hms = c->myhamster->hamster;
  hms_turn(hms,HMS_LEFT);
}


void turn_right(Control *c) {                                          /* Hamster nach links drehen */
  HAMSTER *hms = c->myhamster->hamster;
  hms_turn(hms,HMS_RIGHT);
}




void turn(void (*f)(Control *), Control *c) {                          /* dieser Funktion werden die drehen... Funktionen als Parameter uebergeben */
  (*f)(c);
}


int look_dir(Control *c) {                                             /* in welche Richtung sieht der Hamster gerade? */
  HAMSTER *hms = c->myhamster->hamster;                                
  return hms_dir(hms);                                                 /* Rueckgabewerte laut hamster.h */
}


void scanlook_dir(Control *c) {                                        /* nicht verwendet */

}


void look_north(Control *c) {                                          /* nach Norden sehen */
  turn(turn_north, c);                                                 /* nach Norden drehen */
  look(scan_dir, c);                                                   /* in die Richtung sehen */
}

void look_south(Control *c) {                                          /* wie look_north, nur nach Sueden */
  turn(turn_south, c);
  look(scan_dir, c);
}

void look_east(Control *c) {                                           /* wie look_north, nur nach Osten */
  turn(turn_east, c);
  look(scan_dir, c);
}

void look_west(Control *c) {                                           /* wie look_north, nur nach Westen */
  turn(turn_west, c);
  look(scan_dir, c);
}

void look(void (*f)(Control *), Control *c) {                          /* dieser Funktion werden die schauen... Funktionen uebergeben */
  (*f)(c);
}


void look_field(Control *c) {                                          /* neu betretenes Feld betrachten */
  scan(c);                                                             /* Feld ansehen */
  decide_dir(c); /* entscheiden, in welche Richtung gehen */           /* godir und fromdir ausrechnen */
}


void scan_dir(Control *c) {                                            /* Ausgang in aktueller Richtung bestimmen */
  HAMSTER *hms = c->myhamster->hamster;
  
  if (hms_look(hms) != HMS_WALL) {                                     /* wenn der Hamster nicht auf eine Wand sieht */
    struct MyData *md = c->position.item->node;
    Point p = {md->p.x, md->p.y};
    enum Direction d = none;
    
    switch (look_dir(c)) {                                             /* je nach dem, in welche Richtung der Hamster sieht, werden die Koordinaten des Nachbarfeldes */
    case HMS_NORTH:                                                    /* und der Richtungswert der eigenen Richtungscodierung bestimmt */
      ++p.y;
      d = north;
      break;
    case HMS_SOUTH:
      --p.y;
      d = south;
      break;
    case HMS_EAST:
      ++p.x;
      d = east;
      break;
    case HMS_WEST:
      --p.x;
      d = west;
      break;
    }
    
    if (!search(c, &p)) md->dir |= d;                                  /* nun wird gesucht, ob schon mal ein Feld mit den Koordinaten des Nachbarfeldes betreten wurde */
  }                                                                    /* wenn nein, wird die Richtung als moegliche Richtung aufgenommen */
}


void scan(Control *c) {                                                /* Position, Kornmenge und Ausgaenge bestimmen */
  HAMSTER *hms = c->myhamster->hamster;
  struct MyData *md = c->position.item->node;
  
  md->corn = hms_corn(hms);                                            /* wieviel Korn liegt auf dem Feld? */

  hms_pos(hms, &md->p.x, &md->p.y);                                    /* Koordinaten des Feldes bestimmen */
  
  md->godir=none;                                                      /* wird in decide_dir bestimmt */
  md->fromdir=none;                                                    /*  - " - */

  scan_dir(c);                                                         /* in alle Richtungen, ausser die, aus der der Hamster kommt, sehen */
  turn_right(c);
  scan_dir(c);
  turn_left(c);
  turn_left(c);
  scan_dir(c);
  decide_dir(c);
}


void scan_start(Control *c) {                                          /* wie scan fuer start, es wird aber in alle 4 Richtungen gesehen */
  HAMSTER *hms = c->myhamster->hamster;
  struct MyData *md = c->position.item->node;

  hms_pos(hms, &md->p.x, &md->p.y);
  md->godir=none;                                                      /* wird in decide_dir bestimmt */
  md->fromdir=none;                                                    /*  - " - */
  scan_dir(c);
  turn_right(c);
  scan_dir(c);
  turn_right(c);
  scan_dir(c);
  turn_right(c);
  scan_dir(c);
  decide_dir(c);
}


int search(Control *c, Point *p) {                                     /* suchen, ob Feld mit Position in Liste: 1 - gefunden 0 - nicht gefunden */
  Iterator first = begin(&c->myhamster->list);
  Iterator last = end(&c->myhamster->list);
  struct MyData *md;
  int lookingdir = look_dir(c);

  for( ; first.item != last.item; first = next(first)) {               /* anfangen, die Liste zu durchlaufen */
    md = first.item->node;
    
    if ((md->p.x == p->x) && (md->p.y == p->y)) {                      /* Koordinaten Uebereinstimmung */
      if (lookingdir == HMS_NORTH) {                                   /* moegliche Kreise auftrennen */
	if (md->dir & south) {                                         /* wenn z.B. von Sueden auf das gesuchte Feld gesehen wird, wird der Ausgang Sueden bei */  
	  md->dir ^= south;                                            /* dem gefundenen Feld geloescht */
	}
      }
      if (lookingdir == HMS_SOUTH) {
	if (md->dir & north) {
	  md->dir ^= north;
	}
      }
      if (lookingdir == HMS_WEST) {
	if (md->dir & east) {
	  md->dir ^= east;
	}
      }
      if (lookingdir == HMS_EAST) {
	if (md->dir & west) {
	  md->dir ^= west;
	}
      }                                                                /* moegliche Kreise auftrennen ENDE */
      return 1;                                                        /* Feld gefunden */
    }
  }
  return 0;                                                            /* Feld nicht gefunden */
}


void decide_dir(Control *c) {                                          /* godir und fromdir festlegen */
  struct MyData *md = c->position.item->node;

  Iterator iter = prev(c->position);                                   /* vorhergehendes Feld */
  struct MyData *md2 = iter.item->node;


  if (md->dir & north) {                                               /* wenn moegliche Richtung Norden, godir Norden */
    md->godir = north; 
  } else if (md->dir & east) {                                         /* fuer alle Richtungen */
    md->godir = east;
  } else if (md->dir & west) {
    md->godir = west;
  } else if (md->dir & south) {
    md->godir = south;
  }
  else if (md->dir == none) {
    md->godir = none;
  }

  switch (md2->godir) {                                                /* godir des vorhergehenden Feldes abfragen und dementsprechend fromdir festlegen */
  case north:
    md->fromdir = south;
    break;
  case east:
    md->fromdir = west;
    break;
  case west:
    md->fromdir = east;
    break;
  case south:
    md->fromdir = north;
    break;
  case none:
    md->fromdir = none;
    break;
  }
}




void append_s(Control *c) {                                            /* Startfeld in die Liste eintragen */
  struct MyData *md = malloc(sizeof(struct MyData));
  push_back(&c->myhamster->list, md);                                  /* Feld am Ende der Liste anhaengen */
  md->p.x = 0;                                                         /* initialisieren */
  md->p.y = 0;
  md->dir = 0;
  md->godir = 0;
  md->fromdir = 0;
  md->corn = 0;
}


void append_f(Control *c) {                                            /* neues Element in Liste */

  if (((struct MyData *)c->position.item->node)->godir != none) {      /* wenn vom aktuellen Feld ein neues betreten werden kann, */
    struct MyData *md = malloc(sizeof(struct MyData));
    push_back(&c->myhamster->list, md);                                /* Feld am Ende der Liste anhaengen */
    md->p.x = 0;                                                       /* initialisieren */
    md->p.y = 0;
    md->dir = 0;
    md->godir = 0;
    md->fromdir = 0;
    md->corn = 0;
  }
}


void erase_f(Control *c) {                                             /* Listenelem wo Hamster steht loeschen, und Hamster 1 Feld in Richtung Bau schicken */
  struct MyData *md;

  go(go_back, c);                                                      /* 1 Feld Richtung Bau */
  md = c->position.item->node;
  md->dir ^= md->godir;                                                /* godir da abgearbeitet aus dir loeschen */
  decide_dir(c);
  pop_back (&c->myhamster->list);                                      /* letztes Listenelement loeschen */
}


void erase_s(Control *c) {                                             /* Startfeld loeschen, Liste loeschen */
  erase_List(&c->myhamster->list);
}
